﻿using HistogramImageProcessor.Model;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;

namespace HistogramImageProcessor
{
    public class ImageUtility
    {
        static Stopwatch sw = new Stopwatch();

        public static Bitmap ApplyGrayScale(Bitmap sourceImage)
        {
            sw.Restart();
            byte[] pixelBuffer = GetImageBuffer(sourceImage);
            for (int k = 0; k + 4 < pixelBuffer.Length; k += 4)
            {
                var blue = pixelBuffer[k];
                var green = pixelBuffer[k + 1];
                var red = pixelBuffer[k + 2];
                var gray = red * 0.3 + green * 0.59 + blue * 0.11;
                pixelBuffer[k] = pixelBuffer[k + 1] = pixelBuffer[k + 1] = (byte)Clamp(gray,255,0);
            }
            Bitmap result = SaveToBitmap(pixelBuffer, sourceImage.Width, sourceImage.Height);
            sw.Stop();
            Console.WriteLine("Ellapsed Time(GrayScale): {0} ms", sw.ElapsedMilliseconds);
            return result;


        }
        public static Bitmap ApplyContrastStretching(Bitmap sourceImage)
        {
            sw.Restart();

            byte[] pixelBuffer = GetImageBuffer(sourceImage);
            byte[] redBuffer = new byte[pixelBuffer.Length / 4];
            byte[] greenBuffer = new byte[pixelBuffer.Length / 4];
            byte[] blueBuffer = new byte[pixelBuffer.Length / 4];

            int idx = 0;

            for (int k = 0; k + 4 < pixelBuffer.Length; k += 4)
            {
                blueBuffer[idx] = pixelBuffer[k];
                greenBuffer[idx] = pixelBuffer[k + 1];
                redBuffer[idx] = pixelBuffer[k + 2];
                idx++;
            }

            var rMax = Percentile(redBuffer, 0.95);
            var rMin = Percentile(redBuffer, 0.05);
            var gMax = Percentile(greenBuffer, 0.95);
            var gMin = Percentile(greenBuffer, 0.05);
            var bMax = Percentile(blueBuffer, 0.95);
            var bMin = Percentile(blueBuffer, 0.05);

            var rDelta = 255.0/(rMax - rMin);
            var gDelta = 255.0/(gMax - gMin);
            var bDelta = 255.0/(bMax - bMin);

            idx = 0;

            //Normalize
            for (int k = 0; k + 4 < pixelBuffer.Length; k += 4)
            {
                var blue = (blueBuffer[idx] - bMin) * bDelta;
                var green = (greenBuffer[idx] - gMin) * gDelta;
                var red = (redBuffer[idx] - rMin) * rDelta;

                pixelBuffer[k] = (byte)Clamp(blue, 255, 0);
                pixelBuffer[k+1] = (byte)Clamp(green, 255, 0);
                pixelBuffer[k+2] = (byte)Clamp(red, 255, 0);
                idx++;
            }

            Bitmap result = SaveToBitmap(pixelBuffer, sourceImage.Width, sourceImage.Height);
            sw.Stop();

            Console.WriteLine("Ellapsed Time(Contrast Stretch): {0} ms", sw.ElapsedMilliseconds);
            return result;
        }

        public static Bitmap ApplyContrast(Bitmap sourceImage, double contrast)
        {
            sw.Restart();

            byte[] pixelBuffer = GetImageBuffer(sourceImage);

            var lookup = GetContrastLookup(contrast);

            for (int k = 0; k + 4 < pixelBuffer.Length; k += 4)
            {
                pixelBuffer[k] = lookup[pixelBuffer[k]];            //blue
                pixelBuffer[k + 1] = lookup[pixelBuffer[k + 1]];    //green
                pixelBuffer[k + 2] = lookup[pixelBuffer[k + 2]];    //red
            }

            var resultBitmap = SaveToBitmap(pixelBuffer, sourceImage.Width, sourceImage.Height);

            sw.Stop();
            Console.WriteLine("Ellapsed Time(Contrast): {0} ms", sw.ElapsedMilliseconds);

            return resultBitmap;
        }

        public static Bitmap ApplyBrightness(Bitmap sourceImage, double value)
        {
            sw.Restart();
            byte[] pixelBuffer = GetImageBuffer(sourceImage);

            var lookup = GetBrightnessLookup(value);

            for (int k = 0; k + 4 < pixelBuffer.Length; k += 4)
            {
                pixelBuffer[k] = lookup[pixelBuffer[k]];            //blue
                pixelBuffer[k + 1] = lookup[pixelBuffer[k + 1]];    //green
                pixelBuffer[k + 2] = lookup[pixelBuffer[k + 2]];    //red
            }

            var resultBitmap = SaveToBitmap(pixelBuffer, sourceImage.Width, sourceImage.Height);

            sw.Stop();
            Console.WriteLine("Ellapsed Time(Brightness): {0} ms", sw.ElapsedMilliseconds);

            return resultBitmap;
        }

        public static List<Point> GetGrayLevels(List<ColorRGB> source)
        {
            var data = new List<Point>();
            var ctr = 0;
            foreach(var color in source)
            {
                int gray = (int)(color.Red * 0.3 + color.Green * 0.59 + color.Blue * 0.11);
                data.Add(new Point(ctr++, gray));
            }
            return data;
        }

        public static List<ColorRGB> GetChartData(Bitmap sourceImage)
        {
            sw.Restart();

            byte[] pixelBuffer = GetImageBuffer(sourceImage);

            double blue = 0;
            double green = 0;
            double red = 0;

            var ls = new List<ColorRGB>();

            for (int k = 0; k + 4 < pixelBuffer.Length; k += 4)
            {
                blue = pixelBuffer[k];
                green = pixelBuffer[k + 1];
                red = pixelBuffer[k + 2];
                var color = new ColorRGB(red, green, blue);
                ls.Add(color);
            }
            sw.Stop();
            Console.WriteLine("Ellapsed Time(Load): {0} points in {1} ms",  ls.Count , sw.ElapsedMilliseconds);

            return ls;
        }

        #region Helpers

        private static double Percentile(byte[] seq, double percentile)
        {
            var elements = seq.ToArray();
            Array.Sort(elements);
            double realIndex = percentile * (elements.Length - 1);
            int index = (int)realIndex;
            double frac = realIndex - index;
            if (index + 1 < elements.Length)
                return elements[index] * (1 - frac) + elements[index + 1] * frac;
            else
                return elements[index];
        }

        private static double Clamp(double value, double max, double min)
        {
            return value > max ? max : value < min ? min : value;
        }

        private static byte[] GetImageBuffer(Bitmap sourceImage)
        {
            Bitmap currentImage = (Bitmap)sourceImage.Clone();
            BitmapData sourceData = currentImage.LockBits(new Rectangle(Point.Empty, currentImage.Size),
                ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);

            int bytes = sourceData.Stride * sourceData.Height;
            currentImage.UnlockBits(sourceData);
            byte[] buffer = new byte[bytes];
            Marshal.Copy(sourceData.Scan0, buffer, 0, bytes);
            return buffer;
        }

        private static Bitmap SaveToBitmap(byte[] buffer, int width, int height)
        {
            Bitmap resultBitmap = new Bitmap(width, height);
            BitmapData resultData = resultBitmap.LockBits(new Rectangle(0, 0,resultBitmap.Width, resultBitmap.Height),ImageLockMode.WriteOnly, PixelFormat.Format32bppArgb);
            Marshal.Copy(buffer, 0, resultData.Scan0, buffer.Length);
            resultBitmap.UnlockBits(resultData);
            return resultBitmap;
        }

        private static byte[] GetBrightnessLookup(double brightness)
        {
            var lookup = new byte[256];
            double b = brightness / 100;
            for (int i = 0; i < 256; i++)
            {
                var val = i / 255.0;
                val = (val + val * b) * 255.0;
                val = Clamp(val, 255.0, 1.0);
                lookup[i] = (byte)val;
            }
            return lookup;
        }

        private static byte[] GetContrastLookup(double contrast)
        {
            byte[] lookup = new byte[256];
            double newValue = 0;
            double c = (100.0 + contrast) / 100.0;

            c *= c;

            for (int i = 0; i < 256; i++)
            {
                newValue = ((((i / 255.0) - 0.5) *
                            c) + 0.5) * 255.0;
                newValue = Clamp(newValue, 255, 0);
                lookup[i] = (byte)newValue;
            }
            return lookup;
        }

        #endregion



    }
}
