﻿using C1.Win.Chart;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using HistogramImageProcessor.Model;
using System.Threading.Tasks;
using System.IO;

namespace HistogramImageProcessor
{
    public partial class MainForm : Form
    {
        private List<ColorRGB> _data;
        private Bitmap _originalImage;
        private OpenFileDialog opd;

        public MainForm()
        {
            InitializeComponent();
            _data = new List<ColorRGB>();

            this.ResizeBegin += (s, e) =>
            {
                tableLayoutPanel1.SuspendLayout();
            };

            this.ResizeEnd += (s, e) =>
            {
                tableLayoutPanel1.ResumeLayout();
            };

            InitializeControls();
        }

        private void SetupHistogram(FlexChart chart, string bindingX, string footer, Color fillColor)
        {
            chart.ChartType = C1.Chart.ChartType.Histogram;
            chart.RenderMode = RenderMode.Default;
            chart.DataSource = _data;
            chart.Series.Clear();
            chart.BindingX = bindingX;
            chart.Binding = bindingX;
            var series = new Histogram
            {
                BinWidth = 1,
            };
            series.Style.FillColor = fillColor;
            chart.Series.Add(series);
            chart.ToolTip.Content = "X:{x} Y:{y}";
            chart.AxisX.Min = 0;
            chart.AxisX.Max = 256;
            chart.AxisY.Title = "Pixel count";
            //var font = new Font(chart.Font.FontFamily, 8);
            //chart.AxisY.TitleStyle.Font = font; 
            chart.AxisY.AxisLine = true;
            chart.Footer.Content = footer;
        }

        private void InitializeControls()
        {
            this.SuspendLayout();
            SetupHistogram(flexRed, "Red","Pixel Intensity of Red channel", Color.Red);
            SetupHistogram(flexGreen, "Green", "Pixel Intensity of Green channel",Color.Green);
            SetupHistogram(flexBlue, "Blue","Pixel Intensity of Blue channel", Color.Blue);
            SetupHistogram(flexGray, "Y", "Pixel Intensity of Gray levels", Color.Gray);

            opd = new OpenFileDialog();
            var root = Directory.GetParent(Directory.GetCurrentDirectory()).Parent.FullName;
            opd.InitialDirectory = Path.Combine(root, "Images");
            opd.Filter = "Image files(*.jpg, *.jpeg, *.png) | *.jpg; *.jpeg; *.png";

            var defaultImagePath = Path.Combine(opd.InitialDirectory, "LowContrast3.jpg");
            _originalImage = (Bitmap)Image.FromFile(defaultImagePath);
            Update(_originalImage);
            this.ResumeLayout();
        }

        private void Update(Bitmap image)
        {
            picBox.Image = image;
            var imageSize = picBox.Image.Size;
            var fitSize = picBox.ClientSize;
            picBox.SizeMode = imageSize.Width > fitSize.Width || imageSize.Height > fitSize.Height ?
                PictureBoxSizeMode.StretchImage : PictureBoxSizeMode.CenterImage;
            picBox.Invalidate();

            _data = ImageUtility.GetChartData(image);
            var gray = ImageUtility.GetGrayLevels(_data);
            flexGray.DataSource = gray;
            flexRed.DataSource = _data;
            flexGreen.DataSource = _data;
            flexBlue.DataSource = _data;
        }

        private void OnContrastChange(object sender, EventArgs e)
        {
            if (_originalImage == null)
                return;
            var contrast = trackContrast.Value;
            Bitmap image = null;
            image = ImageUtility.ApplyContrast(_originalImage, contrast);
            Update(image);
        }

        private void OnBrightnessChange(object sender, EventArgs e)
        {
            if (_originalImage == null)
                return;
            var brightness = trackBrightness.Value;
            Bitmap image = null;
            image = ImageUtility.ApplyBrightness(_originalImage, brightness);
            Update(image);
        }

        private void OnAutoContrast(object sender, EventArgs e)
        {
            if (_originalImage == null)
                return;
            trackBrightness.Value = trackContrast.Value = 0;
            Bitmap image = null;
            image = ImageUtility.ApplyContrastStretching(_originalImage);
            Update(image);
        }

        private void OnOpenClick(object sender, EventArgs e)
        {
            if (opd.ShowDialog() == DialogResult.OK)
            {
                _originalImage = (Bitmap)Image.FromFile(opd.FileName);
                Update(_originalImage);
            }
        }

        private void OnSaveClick(object sender, EventArgs e)
        {
            var sfd = new SaveFileDialog();
            sfd.DefaultExt = ".jpeg";
            sfd.ValidateNames = true;
            sfd.Filter = "Image files(*.jpg, *.jpeg, *.png) | *.jpg; *.jpeg; *.png";
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                picBox.Image.Save(sfd.FileName);
            }
        }

        private void OnGrayScaleClick(object sender, EventArgs e)
        {
            if (_originalImage == null)
                return;
            trackBrightness.Value = trackContrast.Value = 0;
            Bitmap image = null;
            image = ImageUtility.ApplyGrayScale(_originalImage);
            Update(image);
        }

        private void OnPanelPaint(object sender, PaintEventArgs e)
        {
            var panel = sender as Panel;
            var pen = new Pen(Color.LightGray);
            var start = new PointF(0, panel.Height - 1);
            var end = new PointF(panel.Width, panel.Height - 1);

            e.Graphics.DrawLine(pen, start, end);

        }
    }
}
